import heapq
from typing import List, Optional

from explorateur.state.storage.base_storage import BaseStorage
from explorateur.state.base_state import BaseState


class PriorityQueue(BaseStorage):
    """  Class representing a priority queue. Root is the smallest element. """

    def __init__(self):
        super().__init__()
        self.storage: List[BaseState] = list()

    def insert(self, state: BaseState):
        heapq.heappush(self.storage, (state.get_objective(), state))

    def remove(self) -> BaseState:
        """ Removes a state from the priority queue."""
        return heapq.heappop(self.storage)[1]  # the first item is the key

    def is_empty(self) -> bool:
        return len(self.storage) == 0

    def size(self) -> int:
        return len(self.storage)

    def contains(self, state: BaseState) -> Optional[BaseState]:
        """ Returns the state if it is in the priority queue, None otherwise."""
        try:
            return self.storage[self.storage.index(state)]
        except ValueError:
            return None



from tests.test_base import BaseTest, EmptyState
from explorateur.state.storage.queue import Queue
from explorateur.state.storage.stack import Stack
from explorateur.state.storage.hash import HashSet
from explorateur.state.storage.priority_queue import PriorityQueue
from tests.test_base import BaseTest, MyState, MyMove


class StorageState(EmptyState):

    def __init__(self, val):
        super().__init__()
        self.val = val

    def get_objective(self):
        if self.val == 2:
            return 1
        else:
            return 10


class StorageTest(BaseTest):

    # can't test for contains() because it takes in a _BaseState
    def test_queue(self):
        q = Queue()
        b1 = StorageState(1)
        b2 = StorageState(2)
        q.insert(b1)
        self.assertTrue(q.contains(b1))
        q.insert(b2)
        res = q.remove()
        self.assertEqual(b1, res)
        self.assertEqual(q.size(), 1)
        q.remove()
        self.assertTrue(q.is_empty())

    def test_stack(self):
        s = Stack()
        b1 = StorageState(1)
        b2 = StorageState(2)
        s.insert(b1)
        self.assertTrue(s.contains(b1))
        s.insert(b2)
        res = s.remove()
        self.assertEqual(b2, res)
        self.assertEqual(s.size(), 1)
        self.assertFalse(s.is_empty())

    def test_hash(self):
        state = MyState({"x": [1, 2], "y": [10, 20], "z": [100, 200]})
        move = MyMove("x", "==", 1)

        h = HashSet()
        self.assertTrue(h.is_empty())

        h.insert(state)
        self.assertTrue(h.contains(state))
        self.assertEqual(h.size(), 1)
        self.assertFalse(h.is_empty())

        d2 = MyState({"x": [1, 2], "y": [10, 20], "z": [100, 200]})
        self.assertEqual(h.contains(d2), d2)

        h.insert(d2)
        self.assertEqual(h.size(), 1)

        x = h.remove()
        self.assertEqual(x, d2)
        self.assertIsNone(h.contains(state))
        self.assertTrue(h.is_empty())
        self.assertEqual(h.size(), 0)

        h.insert(state)
        self.assertEqual(h.size(), 1)


    def test_priority_queue(self):
        # TODO
        pass
        # pq = PriorityQueue()
        # b1 = StorageState(1)
        # b2 = StorageState(2)
        # pq.insert(b1)
        # pq.insert(b2)
        # self.assertEqual(pq.size(), 2)
        # res = pq.remove()
        # self.assertEqual(res, b2)
        # self.assertFalse(pq.is_empty())



import os
from explorateur.utils import Constants
from explorateur.explorateur import Explorateur
from explorateur.search.exploration_type import ExplorationType
from explorateur.search.search_type import SearchType
from tests.test_base import BaseTest, MyState


class InvalidTest(BaseTest):

    def test_no_initial_state(self):
        # Explorateur
        explorer = Explorateur(is_verbose=True)

        # Initial state
        initial_state = MyState({"x": [1, 2], "y": [10, 20], "z": [100, 200]})

        with self.assertRaises(ValueError):
            # Search
            self.search(explorer,
                        initial_state=None,
                        goal_state=None,
                        exploration_type=ExplorationType.DepthFirst(),
                        search_type=SearchType.TreeSearch(),
                        is_solution_path=True,
                        max_depth=100,
                        max_moves=100,
                        max_runtime=100,
                        dot_filename=os.path.join(Constants.TEST_DATA_DIR, "invalid.dot"),
                        is_verbose=True)

    def test_initial_state_type_error(self):
        # Explorateur
        explorer = Explorateur(is_verbose=True)

        # Initial state
        initial_state = MyState({"x": [1, 2], "y": [10, 20], "z": [100, 200]})

        with self.assertRaises(TypeError):
            # Search
            self.search(explorer,
                        initial_state=list(),
                        goal_state=None,
                        exploration_type=ExplorationType.DepthFirst(),
                        search_type=SearchType.TreeSearch(),
                        is_solution_path=True,
                        max_depth=100,
                        max_moves=100,
                        max_runtime=100,
                        dot_filename=os.path.join(Constants.TEST_DATA_DIR, "invalid.dot"),
                        is_verbose=True)

    def test_max_moves_0(self):
        # Explorateur
        explorer = Explorateur(is_verbose=True)

        # Initial state
        initial_state = MyState({"x": [1, 2], "y": [10, 20], "z": [100, 200]})

        with self.assertRaises(ValueError):
            # Search
            self.search(explorer,
                        initial_state,
                        goal_state=None,
                        exploration_type=ExplorationType.DepthFirst(),
                        search_type=SearchType.TreeSearch(),
                        is_solution_path=True,
                        max_depth=100,
                        max_moves=0,
                        max_runtime=100,
                        dot_filename=os.path.join(Constants.TEST_DATA_DIR, "invalid.dot"),
                        is_verbose=True)

    def test_depth_0(self):
        # Explorateur
        explorer = Explorateur(is_verbose=True)

        # Initial state
        initial_state = MyState({"x": [1, 2], "y": [10, 20], "z": [100, 200]})

        with self.assertRaises(ValueError):
            # Search
            self.search(explorer,
                        initial_state,
                        goal_state=None,
                        exploration_type=ExplorationType.DepthFirst(),
                        search_type=SearchType.TreeSearch(),
                        is_solution_path=True,
                        max_depth=0,
                        max_moves=100,
                        max_runtime=100,
                        dot_filename=os.path.join(Constants.TEST_DATA_DIR, "invalid.dot"),
                        is_verbose=True)

    def test_runtime_0(self):
        # Explorateur
        explorer = Explorateur(is_verbose=True)

        # Initial state
        initial_state = MyState({"x": [1, 2], "y": [10, 20], "z": [100, 200]})

        with self.assertRaises(ValueError):
            # Search
            self.search(explorer,
                        initial_state,
                        goal_state=None,
                        exploration_type=ExplorationType.DepthFirst(),
                        search_type=SearchType.TreeSearch(),
                        is_solution_path=True,
                        max_depth=100,
                        max_moves=100,
                        max_runtime=0,
                        dot_filename=os.path.join(Constants.TEST_DATA_DIR, "invalid.dot"),
                        is_verbose=True)


import os
from explorateur.utils import Constants

from tests.test_base import BaseTest, MyState, MyMove
from explorateur import Explorateur, ExplorationType, SearchType


class TreeBestTest(BaseTest):

    def test_tree_best_feasible(self):
        # Explorateur
        explorer = Explorateur(is_verbose=True)

        # Initial state
        # Don't search the whole tree, stop at first feasible
        # Fake a few failing decisions
        initial_state = MyState({"x": [1, 2], "y": [10, 20], "z": [100, 200]},
                                is_exhaustive_search=False)
        # Search
        self.search(explorer,
                    initial_state,
                    goal_state=None,
                    exploration_type=ExplorationType.BestFirst(),
                    search_type=SearchType.TreeSearch(),
                    is_solution_path=True,
                    max_depth=100,
                    max_moves=100,
                    max_runtime=100,
                    dot_filename=os.path.join(Constants.TEST_DATA_DIR, "tree_best_first.dot"),
                    is_verbose=True)

        # Solution label
        ground_truth_solution = {"x": 1, "y": 10, "z": 100}

        self.assertEqual(explorer.num_decisions, 3)
        self.assertEqual(explorer.num_failed_decisions, 0)
        self.assertEqual(len(explorer.solution_path), 4)
        self.assertEqual(explorer.solution_path[-1], initial_state)
        self.assertEqual(explorer.solution_state.var_to_val, ground_truth_solution)
        self.assertEqual(explorer.solution_state.unassigned, [])




import os
from typing import Dict, Union, NamedTuple
from explorateur.search.exploration_type import ExplorationType
from explorateur.search.search_type import SearchType


Num = Union[int, float]
"""Num type is defined as integer or float."""

All_Exploration_Types = Union[ExplorationType.BestFirst, ExplorationType.BreadthFirst, ExplorationType.DepthFirst]
"""All possible exploration types"""

All_Search_Types = Union[SearchType.GraphSearch, SearchType.TreeSearch]
"""All possible search types"""


class Constants(NamedTuple):
    """
    Constant values used by the modules.
    """

    default_seed = 123456
    """The default random seed."""

    # Folder names
    _TEST_DIR_NAME = "tests"
    _DATA_DIR_NAME = "data"

    # Data paths
    _FILE_DIR = os.path.dirname(os.path.abspath(__file__))
    TEST_DATA_DIR = _FILE_DIR + os.sep + ".." + os.sep + _TEST_DIR_NAME + os.sep + _DATA_DIR_NAME

    SUCCESS_NODE_COLOR = "green"
    FAIL_NODE_COLOR = "red"
    LIMIT_NODE_COLOR = "purple"


def run(explorer, args):

    # Search for solutions
    if explorer.search(initial_state=args["initial_state"],
                       goal_state=args["goal_state"],
                       exploration_type=args["exploration_type"],
                       search_type=args["search_type"],
                       is_solution_path=args["is_solution_path"],
                       max_depth=args["max_depth"],
                       max_moves=args["max_moves"],
                       max_runtime=args["max_runtime"],
                       dot_filename=args["dot_filename"]):

        if args["is_verbose"]:
            print("Solution:", explorer.solution_state)
            print("Solution Path:", *explorer.solution_path, sep="\n<-")
    else:
        if args["is_verbose"]:
            print("No solution found!")

    if args["is_verbose"]:
        # Search statistics
        print("Total Decisions:", explorer.num_decisions)
        print("Total Failures:", explorer.num_failed_decisions)
        print("Total Time:", round(explorer.total_time, 3))


def argmax(dictionary: Dict[Num, Num]) -> Num:
    """
    Returns the first key with the maximum value.
    """
    return max(dictionary, key=dictionary.get)


def check_false(expression: bool, exception: Exception) -> None:
    """
    Checks that given expression is false, otherwise raises the given exception.
    """
    if expression:
        raise exception


def check_true(expression: bool, exception: Exception) -> None:
    """
    Checks that given expression is true, otherwise raises the given exception.
    """
    if not expression:
        raise exception


def reset(dictionary: Dict, value) -> None:
    """
    Maps every key to the given value.
    """
    dictionary.update({}.fromkeys(dictionary, value))



from typing import NamedTuple


class ExplorationType(NamedTuple):

    class BestFirst(NamedTuple):
        # _storage: _Storage = _Storage.PriorityQueue()
        pass

    class BreadthFirst(NamedTuple):
        # _storage: _Storage = _Storage.Queue()
        pass

    class DepthFirst(NamedTuple):
        # _storage: _Storage = _Storage.Stack()
        pass



import copy as cp
import time
from typing import Optional, List, Tuple

from explorateur._version import __version__
from explorateur.search.exploration_type import ExplorationType
from explorateur.search.search_type import SearchType
from explorateur.search.transition import Transition
from explorateur.state.base_state import BaseState
from explorateur.state.storage.base_storage import BaseStorage
from explorateur.state.storage.factory import StorageFactory
from explorateur.utils import check_true, All_Exploration_Types, All_Search_Types, Constants

__version__ = __version__


class Explorateur:
    """
        Explorateur is a Python library to conduct State-Space-Search (SSS).
        SSS is a powerful framework for solving problems that require search over a collection of states.

        Explorateur performs generic state-space-search over problem-specific states and moves.
        The user defines the `BaseState` and `BaseMove` and the library drives the search for solutions.

        Given an initial user state, Explorateur performs iterative search moves until a stopping condition is reached:
            - A termination state is found
            - The search space is exhausted
            - Reached max iterations, runtime limit, max depth
            - Optionally, given a goal state, a goal state is encountered.

        The behavior of the overall algorithm is controlled by the Search Strategy and the Exploration Strategy.

        Search Strategy:
            - TreeSearch over open states
            - GraphSearch over open states while also storing the closed states to avoid visiting duplicates.

        Exploration Strategy:
            - BreadthFirst in uninformed fashion
            - DepthFirst in uninformed fashion
            - BestFirst in informed fashion assuming an objective function evaluates the solution quality of a state.

        To use Explorateur, you need to define BaseState and BaseMove, as in the quick start template.
    """

    def __init__(self, is_verbose=False):
        """ Initialize an Explorateur object. """

        # Validate arguments
        Explorateur._validate_args(is_verbose)

        # Set verbosity
        self.is_verbose: bool = is_verbose

        self._log("[Explorateur v." + str(__version__) + "]")

        # Solution state
        self.solution_state: Optional[BaseState] = None
        self.solution_path: Optional[List[BaseState]] = None

        # State collections open and closed (for graph search only)
        self._open: BaseStorage = None
        self.closed: Optional[BaseStorage] = None

        # Statistics
        self._start_time: float = 0
        self.total_time: float = 0
        self.num_decisions: int = 0
        self.num_failed_decisions: int = 0

        # Dot graph text representation of search
        self._dot_text: str = ""
        self._dot_filename: str = ""

    def search(self,
               initial_state: BaseState,
               goal_state: Optional[BaseState] = None,
               exploration_type: All_Exploration_Types = ExplorationType.DepthFirst(),
               search_type: All_Search_Types = SearchType.TreeSearch(),
               is_solution_path: bool = True,
               max_depth: int = 100,
               max_moves: int = 10000,
               max_runtime: int = None,
               dot_filename: str = None) -> bool:
        """
        This function performs search from the initial_state until:
            - a solution found, or
            - goal_state reached if given, or
            - stopping criteria, in terms of number of seconds, iterations, or depth, is reached.

        Arguments:
            - initial_state (BaseState): The initial state where the search will begin.
            - goal_state Optional(BaseState): Optional argument used for graph search to reach termination state.
                                              Default, None.
            - exploration_type Optional(ExplorationType): The exploration method.
                                                          Default, ExplorationType.DepthFirst().
            - search_type Optional(SearchType): The search method to decide whether to store visited states.
                                                Default, SearchType.TreeSearch().
            - is_solution_path (bool): If True, path starting from solution state back to initial_state is returned.
                                       If False, solution state is returned.
                                       Default, True.
            - max_depth Optional(int): Optional argument for the maximum depth to stop search.
                                       Default, 100.
            - max_moves Optional(int): Optional argument for the maximum number of moves to stop search.
                                       Default, 10,000 moves.
            - max_runtime Optional(int): Optional argument for the number of seconds the search should go on for.
                                         Default, None (no limit).
            - dot_file_path (str): Optional argument to write a graph dot representation of search iterations.
                                   It uses logging CRITICAL level to log the dot output
                                   Default, None (no dot file saved).
                                   Example dot graph visualizer: https://dreampuf.github.io/GraphvizOnline/
        Returns:
            - If no solution found or the search hits one of the stopping criteria, returns False.
            - if a solution is found and is_solution_path is set to False, returns the solution state.
            - If a solution is found, and is_solution_path is set to True, returns a list of solution states,
                denoting the path from solution state back to the initial_state.
        """

        # Check arguments
        self._log_start(">>> START SEARCH", max_depth, max_moves, max_runtime)
        Explorateur._validate_search_args(initial_state, goal_state,
                                          exploration_type, search_type, is_solution_path,
                                          max_depth, max_moves, max_runtime, dot_filename)

        # Reset rng, solution_states, collections, dot graph, start time, stats
        self._reset_search(dot_filename)

        # Create storage for open states, and optionally, for closed states in graph search to avoid duplicate visits
        self._open = StorageFactory.create(exploration_type)
        self.closed = StorageFactory.create(search_type)

        # Root node root from the given initial state
        root = cp.deepcopy(initial_state)
        root.id = self.num_decisions

        # Check termination, else expand current state with possible moves
        # as open decisions for execution within depth
        is_terminate, is_solution = self._is_terminate_or_expand(root, goal_state, exploration_type, max_depth)
        if is_solution:
            if is_solution_path:
                self.solution_path = self._get_solution_path()
            return True
        if is_terminate:
            return False

        # START SEARCH
        while not self._open.is_empty():
            self.num_decisions += 1
            self._log("\nDecision " + str(self.num_decisions))
            self._log("Open decisions: " + str(self._open.size()))

            # Pop the copied successor state from open for execution
            successor = self._open.remove()
            current = successor._transition.previous_state
            move = successor._transition.move
            self._log("Current decision: " + str(current))

            # Transition of the current state (there is no transition for initial moves)
            self._log("Current transition: " + str(current._transition))

            # Execute the move on a copy state
            if successor.execute(move):
                self._log("Move is successful.")

                successor.id = self.num_decisions - self.num_failed_decisions
                self._log("Create next transition: " + str(successor._transition) +
                          " from ID: " + str(current.id) + " to ID: " + str(successor.id))

                # Mark the decision as visited, if graph search
                if self.closed:
                    self._log("Mark current decision as visited in closed decisions: " + str(self.closed.size()))
                    self.closed.insert(current)

                # Skip already visited successor, if graph search
                if self.closed and self.closed.contains(successor):
                    self._log("Skip adding decision. It is already visited. " + str(successor))
                    self.num_decisions -= 1
                else:
                    # Create dot node transition
                    self._log_dot(current, move, successor, color="")

                    # Check termination, else expand successor state with possible moves for execution within depth
                    is_terminate, is_solution = self._is_terminate_or_expand(successor, goal_state, exploration_type, max_depth)
                    if is_solution:
                        if is_solution_path:
                            self.solution_path = self._get_solution_path()
                        return True
                    if is_terminate:
                        return False
            else:
                # Skip failed move and infeasible successor, reset failed transition
                successor._transition = None
                self.num_failed_decisions += 1
                self._log("Skip infeasible successor. Num fails: " + str(self.num_failed_decisions))

                # Create dot node transition to a failed node
                self._log_dot(current, move, None, color=Constants.FAIL_NODE_COLOR)

            # Check stopping conditions before next iteration. If hits a limit, color last successor state
            if self._is_search_limit(successor, self._start_time, self.num_decisions, max_runtime, max_moves):
                return False

        # No more open decisions left or limit reach and search finished, save the dot and return False
        self.total_time = time.perf_counter() - self._start_time
        self._log_finish("<<< FINISH SEARCH - FAILURE - No solution! ")
        self._log_dot_file()
        return False

    def _is_terminate_or_expand(self, state, goal_state, exploration_type, max_depth) -> Tuple[bool, bool]:

        is_terminate, is_solution = False, False

        # Check termination condition -- decided by the user state!
        if state.is_terminate(goal_state):
            is_terminate, is_solution = True, True
            self.solution_state = state
            self.total_time = time.perf_counter() - self._start_time
            self._log("Successful termination for state: " + str(state))
            self._log_finish("<<< FINISH SEARCH - SUCCESS - Solution Found!")
            move = state._transition.move if state._transition else None
            self._log_dot(None, move, state, color=Constants.SUCCESS_NODE_COLOR)  # mark it green
            self._log_dot_file()
            return is_terminate, is_solution
        else:
            self._log("Successor is not termination, add alternative moves")

        # If still within max depth bound, insert the successor into open states for exploration
        # Initial state does not have transition, so skip max depth for initial state
        if state._transition:
            if state._transition.depth >= max_depth:
                # Don't terminate the whole search, but don't generate new open nodes
                self._log("Max depth reached, not inserting new open node.")
                self._log_dot(None, state._transition.move, state, color=Constants.LIMIT_NODE_COLOR)
                return is_terminate, is_solution

        # If no termination, add alternative moves to search -- decided by the user state!
        moves = state.get_moves()

        # Reverse moves for depth first search, so the exploration follows user move order
        if isinstance(exploration_type, ExplorationType.DepthFirst):
            moves.reverse()

        # Search for alternatives
        for move in moves:
            self._log("Decision for move: " + str(move))

            # Create a copy successor with next transition
            successor_state = cp.deepcopy(state)
            transition = state._transition
            next_depth = transition.depth + 1 if transition else 1
            next_transition = Transition(previous_state=state, move=move, depth=next_depth)
            successor_state._transition = next_transition

            # Push successor to open storage for execution
            self._log("Add open decision for executing " + str(successor_state._transition.move) + " on state\n" + str(state))
            self._open.insert(successor_state)

        return is_terminate, is_solution

    def _is_search_limit(self, state, start, num_moves, max_runtime, max_moves):
        # Check max_runtime
        stop_cause = None
        current_time = time.perf_counter()
        if max_runtime is not None and current_time - start > max_runtime:
            stop_cause = "Max runtime reached " + str(max_runtime)

        # Check max_moves
        if num_moves >= max_moves:
            stop_cause = "Max number of moves reached " + str(max_moves)

        # If stopped, log, save dot, and return None solution
        if stop_cause:
            self.total_time = current_time - start
            self._log_finish("<<< FINISH SEARCH - STOP - No solution! " + stop_cause)
            self._log_dot(None, state._transition.move, state, color=Constants.LIMIT_NODE_COLOR)  # mark it green
            self._log_dot_file()
            return True
        return False

    def _get_solution_path(self, is_initial_first=False) -> List[BaseState]:
        """
        Returns a list of states representing the order in which the search was performed.

        Args:
            - is_initial_first (bool): Indicates whether to reverse the path (list of states)
            so that the initial state is the first of the list.
            By default, the list has the solution state as the first element of the list
            and traces backwards to the initial state as the last element of the list.

        Returns:
            List[BaseState]: A list of BaseState objects representing the search path.

        """
        # TODO convert recursion to iteration
        state_list = []

        # We can only trace a path if a solution state is found
        if self.solution_state:
            # Start the path from the solution state, and walk backwards
            self._get_path_helper(self.solution_state, state_list)

        return state_list.reverse() if is_initial_first else state_list

    def _get_path_helper(self, state: BaseState, state_list: list):
        """
        This helper recursion traces back explored states explored in reverse order.

        Parameters:
            state (_State): The current state being explored.
            state_list (list): The list to store the explored states.

        Returns:
            list: The list of explored states in reverse order.
        """
        transition = state._transition

        # We reached the initial state, add it and return the path
        if transition is None:
            state_list.append(state)
            return state_list

        # Otherwise, append the base user state to the state list
        state_list.append(state)

        # Continue recursion from the previous internal state
        self._get_path_helper(transition.previous_state, state_list)

    def _log(self, text):
        if self.is_verbose:
            print(text)

    def _log_dot(self, current, move, successor, color):
        if not self._dot_filename:
            return

        if current:
            # If there is no successor, first add a fail node with color so that we can add an edge to it
            if not successor:
                self._dot_text += (str(self.num_decisions) +
                                   " [label=\"Fail: " + str(self.num_failed_decisions) + "\"" +
                                   " shape=triangle style=filled fillcolor="+ str(color) + "];\n")

            # generic nodes with no color
            self._dot_text += "\"" + current.get_dot_label() + "\" -> "
            self._dot_text += "\"" + successor.get_dot_label() + "\"" if successor else str(self.num_decisions)
            self._dot_text += " [label=\"" + move.get_dot_label() + "\"];\n"
        else:
            # decision is taken care of, mark successor with the color
            self._dot_text += "\"" + successor.get_dot_label() + "\" [style=filled fillcolor=" + str(color) + "];\n"

    def _log_dot_file(self):
        if self._dot_filename:
            self._dot_text += "}"
            with open(self._dot_filename, 'w') as dot_file:
                dot_file.write(self._dot_text)

    def _log_start(self, info, max_depth, max_moves, max_runtime):
        self._log("\n" + info +
                  "\nMax Depth: " + str(max_depth) +
                  "\nMax Moves: " + str(max_moves) +
                  "\nMax Time: " + str(max_runtime) + "\n")

    def _log_finish(self, info):
        self._log("\n" + info +
                  "\nTotal Decisions: " + str(self.num_decisions) +
                  "\nTotal Failures: " + str(self.num_failed_decisions) +
                  "\nTotal Time: " + str(round(self.total_time, 3)) + "\n")

    def _reset_search(self, dot_filename):

        # Clean solution states
        self.solution_state = None
        self.solution_path = None

        # Clean state collections
        self._open = None
        self.closed = None

        # Initialize counters
        self._start_time = time.perf_counter()
        self.total_time = 0
        self.num_decisions = 0
        self.num_failed_decisions = 0

        # Dot graph text representation of search, used if dot file given
        self._dot_text = "digraph G {\nspline=line;\n"
        self._dot_filename = dot_filename

    @staticmethod
    def _validate_args(is_verbose) -> None:
        check_true(isinstance(is_verbose, bool),
                   TypeError("is_verbose must be boolean " + str(is_verbose)))

    @staticmethod
    def _validate_search_args(initial_state, goal_state,
                              exploration_type, search_type, is_solution_path,
                              max_depth, max_moves, max_runtime, dot_file_path) -> None:

        check_true(initial_state is not None, ValueError("Initial state cannot be none."))
        check_true(isinstance(initial_state, BaseState),
                   TypeError("Initial state must be BaseState type. Incorrect type: " + str(type(goal_state))))

        if goal_state is not None:
            check_true(isinstance(goal_state, BaseState),
                       TypeError("Goal state must be BaseState type. Incorrect type: " + str(type(goal_state))))

        # Turning off as generic types lead to an error (Python issue)
        # check_true(isinstance(exploration_type, All_Exploration_Types),
        #            TypeError("Exploration type is not allowed " + str(exploration_type)))

        # check_true(isinstance(search_type, All_Search_Types),
        #            TypeError("Search type is not allowed " + str(search_type)))

        check_true(isinstance(is_solution_path, bool),
                   TypeError("is_solution_path must be boolean " + str(is_solution_path)))

        check_true(isinstance(max_depth, int),
                   TypeError("max_depth must be integer number of seconds. Incorrect: " + str(max_depth)))

        check_true(max_depth > 0,
                   ValueError("max_depth must be positive. Incorrect: " + str(max_depth)))

        check_true(isinstance(max_moves, int),
                   TypeError("max_iterations must be integer number of seconds. Incorrect: " + str(max_moves)))

        check_true(max_moves > 0,
                   ValueError("max_moves must be positive. Incorrect: " + str(max_moves)))

        if max_runtime is not None:
            check_true(isinstance(max_runtime, int),
                       TypeError("max_runtime must be integer number of seconds. Incorrect: " + str(max_runtime)))
            check_true(max_runtime > 0,
                       ValueError("max_runtime must be positive. Incorrect: " + str(max_runtime)))

        if dot_file_path is not None:
            check_true(isinstance(dot_file_path, str),
                       TypeError("dot_file_path must be a string. Incorrect type: " + str(type(dot_file_path))))



import unittest
from typing import Dict, List
from explorateur import Explorateur, BaseMove, BaseState, ExplorationType, SearchType


class EmptyMove(BaseMove):

    def __init__(self):
        pass

    def __str__(self) -> str:
        pass


class EmptyState(BaseState):

    def __init__(self):
        super().__init__()

    def get_moves(self):
        pass

    def is_terminate(self, goal_state):
        pass

    def execute(self, move):
        pass

    def __str__(self):
        pass


class MyMove(BaseMove):

    # Define move object
    def __init__(self, var, constraint, val):
        self.var: str = var
        self.constraint: str = constraint
        self.val: int = val

    # String representation, also used for edge labels in DOT graph
    def __str__(self) -> str:
        return str(self.var) + " " + self.constraint + " " + str(self.val)

    # This is required for graph search to check contains on already visited decisions
    def __eq__(self, other):
        if isinstance(other, MyMove):
            return self.var == other.var and \
                self.constraint == other.constraint and \
                self.val == other.val
        return False

    # This is required for graph search when adding closed decision into hashset
    def __hash__(self):
        return hash((self.var, self.constraint, self.val))


class MyState(BaseState):

    # Problem specific state representation
    def __init__(self, var_to_domain, is_exhaustive_search=True, fake_fails=None):
        # IMPORTANT: Make sure to initialize the base state
        super().__init__()

        if fake_fails is None:
            fake_fails = []
        self.var_to_domain: Dict[str, List[int]] = var_to_domain
        self.var_to_val: Dict[str, int] = {}
        self.unassigned: List[str] = list(self.var_to_domain.keys())
        self.is_exhaustive_search = is_exhaustive_search
        self.fake_fails = fake_fails

    # Design branching decisions
    def get_moves(self) -> List[MyMove]:
        # If all vars are assigned, no more moves
        if len(self.unassigned) == 0: return list()

        # Select first unassigned variable and first value in domain
        var = self.unassigned[0]
        val = self.var_to_domain[var][0]

        # Binary branching
        moves = [MyMove(var, "==", val), MyMove(var, "!=", val)]
        print("USER moves [", moves[0], ",", moves[1], "]")
        return moves

    # Check whether a state is the solution/termination
    def is_terminate(self, goal_state=None) -> bool:
        # For demo purposes, no termination, search exhaustively
        if self.is_exhaustive_search: return False
        return len(self.unassigned) == 0

    # Execute the move on the state, return success flag
    def execute(self, move: MyMove) -> bool:
        print("USER Execute:", move)
        var, constraint, val = move.var, move.constraint, move.val

        if move in self.fake_fails:
            print("\t Fake fail [", move, "]")
            return False

        # Propagate EQ and NEQ constraints
        if constraint == "==":
            print("\tEQ assigns", var, "<-", val)
            self.var_to_val[var] = val
            self.var_to_domain[var] = [val]
            self.unassigned.remove(var)
        elif constraint == "!=":
            print("\tNEQ removes", val, "from", var)
            self.var_to_domain[var].remove(val)
            if len(self.var_to_domain[var]) == 0:
                print("USER Execute: FAILS, empty domain var:", var)
                return False
            elif len(self.var_to_domain[var]) == 1:
                self.var_to_val[var] = self.var_to_domain[var][0]
                self.unassigned.remove(var)
                print("\tNEQ fixed", var, "<-", self.var_to_val[var])

        # Successful execution
        print("USER Execute: SUCCESS")
        return True

    # String representation, also used for node labels in DOT graph
    def __str__(self) -> str:
        text = "State ID: " + str(self.id) + "\n"
        text += "Assignment: " + str(self.var_to_val) + "\n"
        text += "Domains: " + str(self.var_to_domain) + "\n"
        text += "Objective: " + str(self.get_objective())
        return text

    # This is required for graph search to check contains on already visited decisions
    def __eq__(self, other):
        if isinstance(other, MyState):
            return self.unassigned == other.unassigned and \
                self.var_to_val == other.var_to_val and \
                self.var_to_domain == other.var_to_domain
        return False

    # This is required for graph search when adding closed decision into hashset
    def __hash__(self):
        return hash((tuple(self.unassigned), tuple(self.var_to_val), tuple(self.var_to_domain)))

    def get_objective(self) -> float:
        objective = 0
        for var in self.var_to_val:
            if var in self.unassigned:
                continue
            else:
                objective += self.var_to_val[var]

        # Return -1, since minimization
        return -1 * objective

class BaseTest(unittest.TestCase):

    def search(self,
               explorer,
               initial_state, goal_state,
               exploration_type, search_type,
               is_solution_path,
               max_depth, max_moves, max_runtime, dot_filename, is_verbose):

        # Search for solutions
        if explorer.search(initial_state,
                           goal_state=goal_state,  # Optional goal state
                           exploration_type=exploration_type,
                           search_type=search_type,
                           is_solution_path=is_solution_path,
                           max_depth=max_depth,
                           max_moves=max_moves,
                           max_runtime=max_runtime,
                           dot_filename=dot_filename):

            if is_verbose:
                print("Solution:", explorer.solution_state)
                print("Solution Path:", *explorer.solution_path, sep="\n<-")
        else:
            if is_verbose:
                print("No solution found!")

        if is_verbose:
            # Search statistics
            print("Total Decisions:", explorer.num_decisions)
            print("Total Failures:", explorer.num_failed_decisions)
            print("Total Time:", round(explorer.total_time, 3))

    def assertListAlmostEqual(self, list1, list2):
        """
        Asserts that floating values in the given lists (almost) equals to each other
        """
        if not isinstance(list1, list):
            list1 = list(list1)

        if not isinstance(list2, list):
            list2 = list(list2)

        self.assertEqual(len(list1), len(list2))

        for index, val in enumerate(list1):
            self.assertAlmostEqual(val, list2[index])



import abc
from typing import Optional

from explorateur.state.base_state import BaseState


class BaseStorage(metaclass=abc.ABCMeta):
    """
    Abstract class for the storage classes.
    """

    @abc.abstractmethod
    def __init__(self):
        """Initializer for the storage class."""

    @abc.abstractmethod
    def insert(self, state: BaseState):
        """Inserts a state into storage."""

    @abc.abstractmethod
    def remove(self) -> BaseState:
        """Removes a state from storage."""

    @abc.abstractmethod
    def is_empty(self) -> bool:
        """Returns True if the storage is empty, False otherwise."""

    @abc.abstractmethod
    def size(self) -> int:
        """Returns the number of elements in the storage."""

    @abc.abstractmethod
    def contains(self, state: BaseState) -> Optional[BaseState]:
        """ Returns the state if it is in the storage, None otherwise."""



import abc


class BaseMove(metaclass=abc.ABCMeta):
    """
    Abstract class for the moves.
    Model your problem as a sequence of states with moves
    encoding information to transition from one state to another.
    """

    @abc.abstractmethod
    def __init__(self):
        """
        Initializes the BaseMove object
        This can contain variable assignments etc.
        """

    @abc.abstractmethod
    def __str__(self) -> str:
        """
        Return a string representation of the move.

        Returns:
            str: A string representation of the move.
        """

    def get_dot_label(self) -> str:
        """
           Return a string label for dot graph edge.

           Returns:
               str: A string label to display in dot graph.
        """
        return str(self)



class Transition:
    """
    Transition from a previous state with a move
    """

    def __init__(self, previous_state: 'BaseState', move: 'BaseMove', depth: int):
        self.previous_state = previous_state
        self.move = move
        self.depth = depth

    def __str__(self):
        return str(self.move) + " depth: " + str(self.depth)



from typing import Set, Optional

from explorateur.state.storage.base_storage import BaseStorage
from explorateur.state.base_state import BaseState


class HashSet(BaseStorage):
    """ Class representing a stack."""

    def __init__(self):
        super().__init__()
        self.storage: Set[BaseState] = set()

    def insert(self, state: BaseState):
        self.storage.add(state)

    def remove(self) -> BaseState:
        return self.storage.pop()

    def is_empty(self) -> bool:
        return len(self.storage) == 0

    def size(self) -> int:
        return len(self.storage)

    def contains(self, state: BaseState) -> Optional[BaseState]:
        """ Returns the state if it is in the hashset, None otherwise."""
        return state if state in self.storage else None



from explorateur.state.storage.queue import Queue
from explorateur.state.storage.stack import Stack
from explorateur.state.storage.priority_queue import PriorityQueue
from explorateur.state.storage.hash import HashSet
from explorateur.state.storage.base_storage import BaseStorage
from explorateur.search.exploration_type import ExplorationType
from explorateur.search.search_type import SearchType
from explorateur.utils import All_Exploration_Types
from typing import Union, Optional


class StorageFactory:
    """ A factory class for creating different types of storage objects."""

    factory = {ExplorationType.BestFirst: PriorityQueue,
               ExplorationType.BreadthFirst: Queue,
               ExplorationType.DepthFirst: Stack,
               SearchType.GraphSearch: HashSet}

    @staticmethod
    def create(storage_type: Union[All_Exploration_Types, SearchType.GraphSearch]) -> Optional[BaseStorage]:
        """
        Create a storage object based on the given storage type.
        If storage type does not exist, returns None.

        Args:
            storage_type (Union[All_Exploration_Types, SearchType.GraphSearch]): The type of storage to create.

        Returns:
            Optional(BaseStorage): The created storage object or None if storage type does not exist.

        """
        storage = StorageFactory.factory.get(type(storage_type))
        return storage() if storage else None



import os
from explorateur.utils import Constants

from tests.test_base import BaseTest, MyState, MyMove
from explorateur import Explorateur, ExplorationType, SearchType
from explorateur.utils import run


class TreeDepthTest(BaseTest):

    def test_tree_depth_feasible(self):
        # Explorateur
        explorer = Explorateur(is_verbose=True)

        # Initial state
        # Don't search the whole tree, stop at first feasible
        # Fake a few failing decisions
        initial_state = MyState({"x": [1, 2], "y": [10, 20], "z": [100, 200]},
                                is_exhaustive_search=False)

        # Arguments
        args = {}
        args["initial_state"] = initial_state
        args["goal_state"] = None
        args["exploration_type"] = ExplorationType.DepthFirst()
        args["search_type"] = SearchType.TreeSearch()
        args["is_solution_path"] = True
        args["max_depth"] = 100
        args["max_moves"] = 100
        args["max_runtime"] = 100
        args["dot_filename"] = os.path.join(Constants.TEST_DATA_DIR, "tree_depth_first.dot")
        args["is_verbose"] = True

        # Run
        run(explorer, args)

        # Solution label
        ground_truth_solution = {"x": 1, "y": 10, "z": 100}
        self.assertEqual(explorer.num_decisions, 3)
        self.assertEqual(explorer.num_failed_decisions, 0)
        self.assertEqual(len(explorer.solution_path), 4)
        self.assertEqual(explorer.solution_path[-1], initial_state)
        self.assertEqual(explorer.solution_state.var_to_val, ground_truth_solution)
        self.assertEqual(explorer.solution_state.unassigned, [])

    def test_tree_depth_feasible_with_fakefails(self):

        # Explorateur
        explorer = Explorateur(is_verbose=True)

        # Initial state
        # Don't search the whole tree, stop at first feasible
        # Fake a few failing decisions
        initial_state = MyState({"x":[1,2], "y":[10, 20], "z":[100, 200]},
                                is_exhaustive_search=False,
                                fake_fails=[MyMove("x", "==", 1), MyMove("z", "==", 100)])

        # Arguments
        args = {}
        args["initial_state"] = initial_state
        args["goal_state"] = None
        args["exploration_type"] = ExplorationType.DepthFirst()
        args["search_type"] = SearchType.TreeSearch()
        args["is_solution_path"] = True
        args["max_depth"] = 100
        args["max_moves"] = 100
        args["max_runtime"] = 100
        args["dot_filename"] = os.path.join(Constants.TEST_DATA_DIR, "tree_depth_first.dot")
        args["is_verbose"] = True

        # Run
        run(explorer, args)

        # Solution label
        ground_truth_solution = {"x": 2, "y": 10, "z": 200}
        self.assertEqual(explorer.num_decisions, 5)
        self.assertEqual(explorer.num_failed_decisions, 2)
        self.assertEqual(len(explorer.solution_path), 4)
        self.assertEqual(explorer.solution_path[-1], initial_state)
        self.assertEqual(explorer.solution_state.var_to_val, ground_truth_solution)
        self.assertEqual(explorer.solution_state.unassigned, [])



import abc
from typing import List, Union, Optional
from explorateur.search.transition import Transition
from explorateur.state.base_move import BaseMove


class BaseState(metaclass=abc.ABCMeta):
    """
    Abstract class for the states.
    Model your problem as a state and transition between states via moves until termination.
    """

    @abc.abstractmethod
    def __init__(self):
        """
        Initializer for state.
            For example, this can store variable and value assignments in the current state.
        """
        self.id: int = -1
        self._transition: Optional[Transition] = None

    @abc.abstractmethod
    def execute(self, move: BaseMove) -> bool:
        """
        Execute the given move on the state and return a boolean to denote successful (valid) execution.

        Parameters:
        move (BaseMove): The move to be executed on the state.

        Returns:
            bool: True if the execution was successful (valid), False otherwise.
        """

    @abc.abstractmethod
    def get_moves(self) -> List[BaseMove]:
        """
        Return an ordered list of moves.
        Moves are alternative search decisions to be explored from this state.

        Returns:
            List[BaseMove]: A list of moves to explore from this state.
        """

    @abc.abstractmethod
    def is_terminate(self, goal_state: Union['BaseState', None]) -> bool:
        """
        Checks if the current state is a termination state.
            If doing tree search, this can check if the state is a solution to the problem.
            If doing graph search, this can check if the state is equivalent to the given end state.

        Parameters:
            goal_state: The end state to compare with (optional).

        Returns:
            bool: True if the current state is a termination state, False otherwise.
        """

    @abc.abstractmethod
    def __str__(self) -> str:
        """
        Return a string representation of the state.

        Returns:
            str: A string representation of the state.
        """

    def get_dot_label(self) -> str:
        """
       Return a string label for dot graph node.

       Returns:
           str: A string label to display in dot graph.
        """
        return str(self)

    def get_objective(self) -> float:
        """
        Return the objective value of the state.

        Best-First Search requires an evaluation of states for ranking.
        By default, Best-First Search is set to minimization.
        To maximize, multiply your objective by -1.  

        Note, if this function returns the same value for two states
        there will be an error because there is no comparator between abstract Base States.

        Returns:
            float: The objective function value for the state.
        """



import os
from explorateur.utils import Constants

from tests.test_base import BaseTest, MyState
from explorateur import Explorateur, ExplorationType, SearchType
from explorateur.utils import run


class UsageExampleTest(BaseTest):

    def test_usage_example(self):

        # Explorateur
        explorer = Explorateur(is_verbose=True)

        # Initial state
        initial_state = MyState({"x": [1,2], "y": [10, 20], "z": [100, 200]})

        # Arguments
        args = {}
        args["initial_state"] = initial_state
        args["goal_state"] = None
        args["exploration_type"] = ExplorationType.DepthFirst()
        args["search_type"] = SearchType.TreeSearch()
        args["is_solution_path"] = True
        args["max_depth"] = 100
        args["max_moves"] = 100
        args["max_runtime"] = 100
        args["dot_filename"] = os.path.join(Constants.TEST_DATA_DIR, "usage_example.dot")
        args["is_verbose"] = True

        # Run
        run(explorer, args)
        self.assertEqual(explorer.num_decisions, 14)
        self.assertEqual(explorer.num_failed_decisions, 0)



from explorateur._version import __version__
from explorateur.explorateur import Explorateur
from explorateur.search.exploration_type import ExplorationType
from explorateur.search.search_type import SearchType
from explorateur.state.base_state import BaseState
from explorateur.state.base_move import BaseMove

